/*
    6-19-2007
    Copyright Spark Fun Electronics 2007
    Nathan Seidle
    nathan at sparkfun.com
    
    Simon Game ported for the ATmega8
	
    Generates random sequence, plays music, and displays button lights.

    Simon tones from Wikapedia
    
    * A (red, upper left) - 440Hz - 2.272ms - 1.136ms pulse
    * a (green, upper right, an octave higher than the upper right) - 880Hz - 1.136ms - 0.568ms pulse
    * D (blue, lower left, a perfect fourth higher than the upper left) - 587.33Hz - 1.702ms - 0.851ms pulse
    * G (yellow, lower right, a perfect fourth higher than the lower left) - 784Hz - 1.276ms - 0.638ms pulse

    The tones are close, but probably off a bit, but they sound all right.
    
    Timer rolls over at 256 and is reset to 131 (Timer 2 ISR) so 125 possible numbers
    131 + 31 = 162 will be button 1
    162 + 31 = 193 button 2
    193 + 31 = 224 button 3
    224 + 31 = 255 button 4

	Prescalar of 1024
	Clock = 16MHz
	15,625 clicks per second
	64us per click

    So each button has an equal chance of being added to the game sequence
    Each timer2 click is every 64us (16MHz / 1024) so the button possibility 
    changes every 2ms or 500 times a second
    Not random, but no one can play the same way to within 2ms!
    The sequence changes *every* game
	
	Original Fuses : avrdude -p atmega8 -P lpt1 -c stk200 -U lfuse:w:0xE1:m -U hfuse:w:0xD9:m
	16MHz Fuses : avrdude -p atmega8 -P lpt1 -c stk200 -U lfuse:w:0xEE:m -U hfuse:w:0xC9:m
*/

#include <avr/io.h>
#include <avr/interrupt.h>

#define sbi(var, mask)   ((var) |= (uint8_t)(1 << mask))
#define cbi(var, mask)   ((var) &= (uint8_t)~(1 << mask))

#define BLU_LED	2
#define YLW_LED	3
#define RED_LED	4
#define GRN_LED	5

#define LED_OFF	0
#define LED_ON	1

#define BUTTON0	4
#define BUTTON1	5
#define BUTTON2	6
#define BUTTON3	7

#define BUZZ0	0
#define BUZZ1	1

uint16_t random_byte;

uint8_t game_string[32];
uint8_t game_string_position;

volatile uint16_t global_clock;
volatile uint8_t flag;

//Define functions
//======================
void toner(uint8_t, uint8_t);
void add_to_string(void);
void play_string(void);
void play_loser(void);
void play_winner(void);

void ioinit(void);      //Initializes IO
void delay_ms(uint16_t x); //General purpose delay
void delay_us(uint16_t x);
//======================

ISR (SIG_OVERFLOW2) 
{
    global_clock++;

	//Prescalar of 1024
	//Clock = 16MHz
	//15,625 clicks per second
	//64us per click	
    if(global_clock == 125) //(125 * (125 * .000064) = 1.000s
    {
        flag = 1; //One second has hit
        global_clock = 0;
    }
    
    TCNT2 = 131; //256 - 125 = 131 - Preload timer 2 for 125 clicks. Should be 8ms per ISR call
}

int main (void)
{
    uint8_t choice;
    uint8_t current_pos;
    uint16_t time_limit;
	uint32_t test[10][10];
	
	
	test[3][2] = 32;

    ioinit(); //Setup IO pins and defaults
	
	//Test routines
    /*
	toner('1', 100);
    delay_ms(100);
    toner('2', 100);
    delay_ms(100);
    toner('3', 100);
    delay_ms(100);
    toner('4', 100);
    delay_ms(100);
	*/

BEGIN_GAME:

    //Display fancy LED pattern waiting for two buttons to be pressed
    //Wait for user to begin game
    while(1)
    {
        sbi(PORTC, RED_LED);
        cbi(PORTC, BLU_LED);
        cbi(PORTC, GRN_LED);
        cbi(PORTC, YLW_LED);
        delay_ms(100);
        if( (PIND & 0xF0 ) != 0xF0 ) break;

        cbi(PORTC, RED_LED);
        sbi(PORTC, BLU_LED);
        cbi(PORTC, GRN_LED);
        cbi(PORTC, YLW_LED);
        delay_ms(100);
        if( (PIND & 0xF0) != 0xF0 ) break;

        cbi(PORTC, RED_LED);
        cbi(PORTC, BLU_LED);
        sbi(PORTC, GRN_LED);
        cbi(PORTC, YLW_LED);
        delay_ms(100);
        if( (PIND & 0xF0) != 0xF0 ) break;

        cbi(PORTC, RED_LED);
        cbi(PORTC, BLU_LED);
        cbi(PORTC, GRN_LED);
        sbi(PORTC, YLW_LED);
        delay_ms(100);
        if( (PIND & 0xF0) != 0xF0 ) break;
    }
    
    //Indicate the start of game play
	PORTC = 0b00111100;
    delay_ms(1000);
	PORTC = 0b00000000;
    delay_ms(250);

    game_string_position = 0; //Start new game

    while(1)
    {
        add_to_string(); //Add the first button to the string
        play_string(); //Play the current contents of the game string back for the player
    
        //Wait for user to input buttons until they mess up, reach the end of the current string, or time out
        for(current_pos = 0 ; current_pos < game_string_position ; current_pos++)
        {
            //Button scanning
            //=======================================
            time_limit = 0; //Run timer while we wait for the user to push a button

            while(1)
            {
				if( (PIND & 0xF0) != 0xF0 ) break;

                delay_ms(1);

                time_limit++; //3 second time limit
                if (time_limit > 3000)
                {
                    play_loser(); //Play anoying loser tones
                    goto BEGIN_GAME;
                }
            }

            //Wait for user to release button
            choice = 0;
            while(choice == 0)
            {
                if( !(PIND & (1 << BUTTON0)) ) choice = '3'; //Lower left button
                if( !(PIND & (1 << BUTTON1)) ) choice = '4'; //Lower right
                if( !(PIND & (1 << BUTTON2)) ) choice = '1'; //Upper left
                if( !(PIND & (1 << BUTTON3)) ) choice = '2'; //Upper right
                
                toner(choice, 50); //Fire the button and play the button tone
            }
            //=======================================
        
            if (choice != game_string[current_pos])
            {
                play_loser(); //Play anoying loser tones
                goto BEGIN_GAME;
            }

        }//End user input loop
        
        //If user reaches the game length of X, they win!
        if (current_pos == 13)
        {
            play_winner(); //Play winner tones
            goto BEGIN_GAME;
        }
        
        //Otherwise, we need to wait just a hair before we play back the last string
        delay_ms(1000);
    }
   
    return(0);
}

void ioinit (void)
{
    //1 = output, 0 = input
    DDRB = 0b11111111; //All outputs
    DDRC = 0b11111111; //4-7 LED outputs, All outputs
    DDRD = 0b00001111; //4-7 button inputs

    //Init timer 2
    ASSR = 0;
    TCCR2 = (1<<CS22)|(1<<CS21)|(1<<CS20); //Set Prescaler to 1024. CS22=1, CS21=1,CS20=1
    TIMSK = (1<<TOIE2); //Enable Timer 2 Interrupt
    
    sei();
	
}

//Plays the loser sounds
void play_loser(void)
{
    //printf("You failed!\n", 0);
    
    sbi(PORTC, RED_LED);
    sbi(PORTC, GRN_LED);
    toner('0', 255);

    sbi(PORTC, BLU_LED);
    sbi(PORTC, YLW_LED);
    toner('0', 255);

    sbi(PORTC, RED_LED);
    sbi(PORTC, GRN_LED);
    toner('0', 255);
}

//Plays the winner sounds
void play_winner(void)
{
    uint8_t x, y, z;

    //printf("You win!\n", 0);

    sbi(PORTC, RED_LED);
    sbi(PORTC, YLW_LED);
    sbi(PORTC, BLU_LED);
    sbi(PORTC, GRN_LED);
    
    for(z = 0 ; z < 4 ; z++)
    {
        if(z == 0 || z == 2)
		{
			cbi(PORTC, RED_LED);
			cbi(PORTC, YLW_LED);
			sbi(PORTC, BLU_LED);
			sbi(PORTC, GRN_LED);
		}
		else
		{
			sbi(PORTC, RED_LED);
			sbi(PORTC, YLW_LED);
			cbi(PORTC, BLU_LED);
			cbi(PORTC, GRN_LED);
		}

        for(x = 250 ; x > 70 ; x--)
        {
            for(y = 0 ; y < 3 ; y++)
            {
                //Toggle the buzzer at various speeds
				sbi(PORTC, BUZZ1);
				cbi(PORTC, BUZZ0);
                
                delay_us(x);
                
				cbi(PORTC, BUZZ1);
				sbi(PORTC, BUZZ0);
                
                delay_us(x);
            }
        }
    }
            
}

//Plays the current contents of the game string
void play_string(void)
{
    uint8_t string_pos;
    uint8_t button;
    
    for(string_pos = 0 ; string_pos < game_string_position ; string_pos++)
    {
        button = game_string[string_pos];
        toner(button, 100);
        delay_ms(150);
    }
    
}

//Adds a new random button to the game sequence based on the current timer elapsed
void add_to_string(void)
{
    uint8_t random_timer;
    uint8_t new_button;

	random_timer = TCNT2;

    //Timer rolls over at 256 and is reset to 131 (Timer 2 ISR) so 125 possible numbers
    //131 + 31 = 162 will be button 1
    //162 + 31 = 193 button 2
    //193 + 31 = 224 button 3
    //224 + 31 = 255 button 4

	//Prescalar of 1024
	//Clock = 16MHz
	//15,625 clicks per second
	//64us per click

    //So each button has an equal chance of being added to the game sequence
    //Each timer2 click is every 64us (16MHz / 1024) so the button possibility 
    //changes every 2ms or 500 times a second
    //Not random, but no one can play the same way to within 2ms!
    //The sequence changes *every* game
    
    if(random_timer < 162) 
        new_button = '1';
    else if(random_timer < 193) 
        new_button = '2';
    else if(random_timer < 224) 
        new_button = '3';
    else
        new_button = '4';
        
    game_string[game_string_position] = new_button;
    game_string_position++;
    
}

//Tone generator
void toner(uint8_t tone, uint8_t buzz_length)
{
    uint16_t buzz_delay;
    uint16_t loop_length;
    
    switch (tone)
    {
        case 'z' :
        case '0' :
            //button = ; //Failure Light up all LEDS
            buzz_delay = 300; //1500;
            loop_length = 132;//33;
            break;

        case 'A' :
        case '1' :
            buzz_delay = 227;//1136; //Upper left, Red
            loop_length = 176;//44;
            sbi(PORTC, RED_LED);
            break;

        case 'a' :
        case '2' :
            buzz_delay = 113;//568; //Upper right, Green
            loop_length = 352;//88;
            sbi(PORTC, GRN_LED);
            break;

        case 'D' :
        case '3' :
            buzz_delay = 170;//851; //Lower left, Blue
            loop_length = 236;//59;
            sbi(PORTC, BLU_LED);
            break;

        case 'G' :
        case '4' :
            buzz_delay = 127;//638; //Lower right, Yellow
            loop_length = 312;//78;
            sbi(PORTC, YLW_LED);
            break;

        //Special noises without LEDs
        case '5' :
            buzz_delay = 150;//300; //Special noise
            loop_length = 300;//167;
            break;

        case '6' :
            buzz_delay = 160;//300; //Special noise
            loop_length = 300;//167;
            break;
    }
    
    //Run buzzer for ~250ms?
    for( ; buzz_length > 0 ; buzz_length--)
    {
        for( ; loop_length > 0 ; loop_length--)
        {
            //Toggle the buzzer at various speeds
            sbi(PORTC, BUZZ1);
            cbi(PORTC, BUZZ0);
            
            delay_us(buzz_delay);
            
            cbi(PORTC, BUZZ1);
            sbi(PORTC, BUZZ0);
            
            delay_us(buzz_delay);
        }
    }

    //Turn off all LEDs
    cbi(PORTC, RED_LED);
    cbi(PORTC, BLU_LED);
    cbi(PORTC, YLW_LED);
    cbi(PORTC, GRN_LED);
       
}

//Really short delay
void delay_us(uint16_t x)
{
  uint8_t y, z;
  for ( ; x > 0 ; x--){
    for ( y = 0 ; y < 5 ; y++){
      for ( z = 0 ; z < 6 ; z++){
        asm volatile ("nop");
      }
    }
  }
}

//General short delays
void delay_ms(uint16_t x)
{
  uint8_t y, z;
  for ( ; x > 0 ; x--){
    for ( y = 0 ; y < 55 ; y++){
      for ( z = 0 ; z < 61 ; z++){
        asm volatile ("nop");
      }
    }
  }
}
